from pathlib import Path
from typing import List, Optional, Protocol
from fsspec import AbstractFileSystem
from fsspec.implementations.local import LocalFileSystem
import numpy as np
import pandas as pd

from physrisk.kernel.assets import RealEstateAsset
from physrisk.kernel.hazards import (
    CoastalInundation,
    PluvialInundation,
    RiverineInundation,
)
from physrisk.kernel.impact_distrib import ImpactType
from physrisk.vulnerability_models.config_based_impact_curves import (
    VulnerabilityConfigItem,
    config_items_to_csv,
)


def build_all_config(source_dir: Optional[Path] = None):
    """Build all of the configuration."""
    config_items: list[VulnerabilityConfigItem] = []
    source_dir = vulnerability_onboarding_dir() if source_dir is None else source_dir
    config_builders = [ConfigBuilderJRCFlood(source_dir=source_dir)]
    for builder in config_builders:
        config_items += builder.build_config()
    config_items_to_csv(
        config_items, vulnerability_config_dir() / "candidate_vulnerability_config.csv"
    )


def vulnerability_onboarding_dir():
    """Path of the vulnerability on-boarding directory."""
    path = (
        Path(__file__).parents[4]
        / "docs"
        / "user_guide"
        / "vulnerability"
        / "vulnerability_functions"
    )
    return path


def vulnerability_config_dir():
    """Path of the vulnerability configuration directory."""
    path = Path(__file__).parents[2] / "data" / "static" / "vulnerability"
    return path


class ConfigBuilder(Protocol):
    def build_config(self) -> List[VulnerabilityConfigItem]: ...


class ConfigBuilderBase:
    def __init__(self, source_dir: Path, fs: Optional[AbstractFileSystem] = None):
        """_summary_

        Args:
            source_dir (str): Full path to directory containing any source files.
                Defaults to None, in which case the default on-boarding directory in the
                repo is used.
            fs (Optional[AbstractFileSystem], optional): Instance of AbstractFileSystem to use to
                access source_dir e.g. S3FileSystem.
                Defaults to None, in which case a LocalFileSystem is assumed.
        """
        self.fs = fs if fs is not None else LocalFileSystem()
        self.source_dir = source_dir


class ConfigBuilderJRCFlood(ConfigBuilderBase, ConfigBuilder):
    def build_config(self):
        """Create vulnerability config for flood based on the EU JRC global depth-damage functions.
        This replaces the notebook implementation and sets the pattern for code-based generation of the config.
        """

        # raw data is actually a reformatting of the original Excel spreadsheet to facilitate loading.
        raw_data = self.source_dir / "inundation_jrc" / "raw.csv"
        df = pd.read_csv(raw_data)

        # consistent with physrisk continent definition
        location_mappings = {
            "Europe": "Europe",
            "North America": "North America",
            "Central & South America": "South America",
            "Asia": "Asia",
            "Africa": "Africa",
            "Oceania": "Oceania",
            "Global": "Global",
        }
        type_mappings = {
            "Residential buildings": "Buildings/Residential",
            "Commercial buildings": "Buildings/Commercial",
            "Industrial buildings": "Buildings/Industrial",
        }

        config_items: List[VulnerabilityConfigItem] = []
        for mapping in type_mappings:
            type_df = df[df["Type"] == mapping]
            flood_depth = type_df["Flood depth [m]"].to_numpy()
            for location in location_mappings:
                # whether zero depth is considered really zero or a flood event with smallest depth
                zero_as_minimum = True if location == "North America" else False
                # for North America, the 0 depth damage is for flooding of any depth. We consider that a 1 cm inundation.
                depth = (
                    np.concatenate([[0, 0.01], flood_depth[1:]])
                    if zero_as_minimum
                    else flood_depth
                )

                mean = type_df[location + "_Mean"].to_numpy()
                std = type_df[location + "_Std"].to_numpy()
                mean = np.concatenate([[0], mean]) if zero_as_minimum else mean
                std = np.concatenate([[0], std]) if zero_as_minimum else std
                if np.any(np.isnan(mean)):
                    mean = []
                if np.any(np.isnan(std)):
                    std = []
                if not any(mean):
                    continue
                config_items.append(
                    VulnerabilityConfigItem(
                        hazard_class=",".join(
                            [
                                CoastalInundation.__name__,
                                PluvialInundation.__name__,
                                RiverineInundation.__name__,
                            ]
                        ),
                        asset_class=RealEstateAsset.__name__,
                        asset_identifier=f"type={type_mappings[mapping]},location={location_mappings[location]}",
                        indicator_id="flood_depth",
                        indicator_units="metres",
                        impact_id=str(ImpactType.damage.name),
                        curve_type="indicator/piecewise_linear",
                        points_x=list(depth),
                        points_y=list(mean),
                        # points_z=std, # omit for now: until implemented
                    )
                )
        return config_items
